from typing import Union

import neurokit2 as nk
import numpy as np
import pandas as pd
from tpcp import Parameter

from biopsykit.signals._dtypes import assert_sample_columns_int
from biopsykit.signals.ecg.segmentation._base_segmentation import BaseHeartbeatSegmentation

__all__ = ["HeartbeatSegmentationNeurokit"]

from biopsykit.utils._datatype_validation_helper import _assert_has_columns
from biopsykit.utils.array_handling import sanitize_input_dataframe_1d


class HeartbeatSegmentationNeurokit(BaseHeartbeatSegmentation):
    """Segments ECG signal into heartbeats based on Neurokit's delineate function."""

    _action_methods = "extract"

    # input parameters
    variable_length: Parameter[bool]
    start_factor: Parameter[float]
    r_peak_detection_method: Parameter[str]

    # result
    heartbeat_list_: pd.DataFrame

    def __init__(
        self,
        *,
        variable_length: bool = True,
        start_factor: float = 0.35,
        r_peak_detection_method: str = "neurokit",
    ):
        """Initialize new HeartBeatExtraction algorithm instance.

        Parameters
        ----------
        variable_length : bool, optional
            ``True`` if extracted heartbeats should have variable length (depending on the current RR-interval) or
            ``False`` if extracted heartbeats should have fixed length (same length for all heartbeats, depending
            on the mean heartrate of the complete signal, 35% of mean heartrate in seconds before R-peak and 50%
            after r_peak, see :func:`neurokit2.ecg_segment` for details).
            For variable length heartbeats, the start of the next heartbeat follows directly after end of last
            (ends exclusive); For fixed length heartbeats, there might be spaces between heartbeat borders, or they
            might overlap. Default: ``True``
        start_factor : float, optional
            only needed if ``variable_length=True``. If ``variable_length=True``, this parameter defines where the
            start border between heartbeats is set depending on the RR-interval to previous heartbeat.
            For example, ``start_factor=0.35`` means that the beat start is set at 35% of current RR-distance before the
            R-peak of the beat
        r_peak_detection_method : str, optional
            Method to detect R-peaks that is passed to :func:`neurokit2.ecg_peaks`. Default: "neurokit"

        """
        self.variable_length = variable_length
        self.start_factor = start_factor
        self.r_peak_detection_method = r_peak_detection_method

    # @make_action_safe
    def extract(
        self,
        *,
        ecg: Union[pd.Series, pd.DataFrame],
        sampling_rate_hz: int,
    ):
        """Segments ecg signal into heartbeats, extract start, end, r-peak of each heartbeat.

        The function uses R-peak detection to segment the ECG signal into heartbeats. The start of each heartbeat is
        determined based on the R-peak and the current RR-interval.

        The results are saved in the ``heartbeat_list_`` attribute.

        Parameters
        ----------
        ecg : :class:`~pandas.Series` or :class:`~pandas.DataFrame`
            ECG signal
        sampling_rate_hz : int
            Sampling rate of ECG signal in hz

        Returns
        -------
            self

        """
        ecg = sanitize_input_dataframe_1d(ecg, column="ECG")
        if ecg.empty:
            raise ValueError("Input data is empty!")
        _assert_has_columns(ecg, [["ECG"]])

        _, r_peaks = nk.ecg_peaks(ecg, sampling_rate=sampling_rate_hz, method=self.r_peak_detection_method)
        r_peaks = r_peaks["ECG_R_Peaks"]

        heartbeats = pd.DataFrame(
            index=np.arange(0, len(r_peaks)),
        )
        heartbeats = heartbeats.assign(r_peak_sample=r_peaks)

        # save RR-interval to successive heartbeat
        heartbeats = heartbeats.assign(rr_interval_sample=np.abs(heartbeats["r_peak_sample"].diff(periods=-1)))

        if self.variable_length:
            # split ecg signal into heartbeats with varying length
            rr_interval_samples = heartbeats["r_peak_sample"].diff()

            # calculate start of each heartbeat based on corresponding R-peak and current RR-interval
            beat_starts = heartbeats["r_peak_sample"] - self.start_factor * rr_interval_samples

            # extrapolate first beats start based on RR-interval of next beat
            first_beat_start = heartbeats["r_peak_sample"].iloc[0] - self.start_factor * rr_interval_samples.iloc[1]
            if first_beat_start >= 0:
                beat_starts.iloc[0] = first_beat_start
            else:
                beat_starts = beat_starts.iloc[1:].reset_index(drop=True)  # drop row if heartbeat is incomplete
                heartbeats = heartbeats.iloc[1:].reset_index(drop=True)
            beat_starts = round(beat_starts).astype(int)

            # calculate beat ends (last beat ends 1 sample before next starts, end is exclusive)
            beat_ends = beat_starts.shift(-1)  # end is exclusive

            # extrapolate last beats end based on RR-interval of previous beat
            last_beat_end = round(
                heartbeats["r_peak_sample"].iloc[-1] + (1 - self.start_factor) * rr_interval_samples.iloc[-1]
            )

            if last_beat_end < len(ecg):
                beat_ends.iloc[-1] = last_beat_end
            else:
                # drop the last beat if it is incomplete
                heartbeats = heartbeats.iloc[:-1]
                beat_ends = beat_ends.iloc[:-1]
                beat_starts = beat_starts.iloc[:-1]
            beat_ends = beat_ends.astype(int)

            # extract time of each beat's start
            beat_starts_time = ecg.iloc[beat_starts].index
            heartbeats = heartbeats.assign(start_sample=beat_starts, end_sample=beat_ends, start_time=beat_starts_time)

        else:
            # split ecg signal into heartbeats with fixed length
            heartbeat_segments = nk.ecg_segment(ecg, rpeaks=r_peaks, sampling_rate=sampling_rate_hz, show=False)

            heartbeat_segments_new = {int(k) - 1: v for k, v in heartbeat_segments.items()}
            heartbeat_segments_new = pd.concat(heartbeat_segments_new, names=["heartbeat_id"])

            heartbeat_segments_new = heartbeat_segments_new.groupby("heartbeat_id").agg(
                start_sample=("Index", "first"),
                end_sample=("Index", "last"),
                start_time=("Index", lambda s: ecg.index[s.iloc[0]]),
            )
            # fill the empty columns of heartbeats with the start, end, and r-peak of heartbeat_segments_new
            heartbeats = heartbeats.join(heartbeat_segments_new)

        # check if R-peak occurs between corresponding start and end
        check = heartbeats.apply(lambda x: x["start_sample"] < x["r_peak_sample"] < x["end_sample"], axis=1)
        if len(check.loc[~check]) > 0:
            raise ValueError(
                f"Start/end/R-peak position of heartbeat {list(check.loc[check is False].index)} could be incorrect!"
            )

        # ensure that index is Int64Index (not RangeIndex) because some neurokit functions won't work  with RangeIndex
        heartbeats.index = list(heartbeats.index)
        heartbeats.index.name = "heartbeat_id"

        heartbeats = heartbeats.assign(rr_interval_ms=heartbeats["rr_interval_sample"] / sampling_rate_hz * 1000)

        # ensure correct column order
        heartbeats = heartbeats[
            ["start_time", "start_sample", "end_sample", "r_peak_sample", "rr_interval_sample", "rr_interval_ms"]
        ]

        heartbeats = heartbeats.convert_dtypes(infer_objects=True)
        # assert that columns with "_sample" in the end are of type int
        assert_sample_columns_int(heartbeats)

        self.heartbeat_list_ = heartbeats
        return self



from typing import Literal

import pandas as pd
from tpcp import Algorithm
from typing_extensions import Self

__all__ = ["PepExtraction"]

NEGATIVE_PEP_HANDLING = Literal["nan", "zero", "keep"]


class PepExtraction(Algorithm):

    _action_methods = "extract"

    handle_negative_pep: NEGATIVE_PEP_HANDLING
    pep_results_: pd.DataFrame

    def __init__(self, handle_negative_pep: NEGATIVE_PEP_HANDLING = "nan") -> None:
        self.handle_negative_pep = handle_negative_pep

    def extract(
        self,
        *,
        heartbeats: pd.DataFrame,
        q_peak_samples: pd.DataFrame,
        b_point_samples: pd.DataFrame,
        sampling_rate_hz: float,
    ) -> Self:
        """Compute PEP from Q-peak samples and B-point locations.

        Args:
            heartbeats:
                Heartbeat locations as DataFrame
            q_peak_samples:
                ECG signal as DataFrame
            b_point_samples:
                ICG signal as DataFrame
            sampling_rate_hz:
                Sampling rate of the signals in Hz

        Returns
        -------
            self
        """
        # do something
        pep_results = pd.DataFrame(index=heartbeats.index)

        pep_results = pep_results.assign(
            heartbeat_start_time=heartbeats["start_time"],
            heartbeat_start_sample=pd.to_numeric(heartbeats["start_sample"]),
            heartbeat_end_sample=pd.to_numeric(heartbeats["end_sample"]),
            r_peak_sample=pd.to_numeric(heartbeats["r_peak_sample"]),
            rr_interval_sample=pd.to_numeric(heartbeats["rr_interval_sample"]),
            rr_interval_ms=pd.to_numeric(heartbeats["rr_interval_sample"] / sampling_rate_hz * 1000),
            heart_rate_bpm=pd.to_numeric(60 / (heartbeats["rr_interval_sample"] / sampling_rate_hz)),
            q_peak_sample=pd.to_numeric(q_peak_samples["q_peak_sample"]),
            b_point_sample=pd.to_numeric(b_point_samples["b_point_sample"]),
            pep_sample=pd.to_numeric(b_point_samples["b_point_sample"] - q_peak_samples["q_peak_sample"]),
        )

        pep_results = pep_results.assign(
            pep_ms=pep_results["pep_sample"] / sampling_rate_hz * 1000,
        )
        pep_results = self._add_invalid_pep_reason(pep_results, q_peak_samples, b_point_samples)
        pep_results = pep_results.convert_dtypes(infer_objects=True)

        self.pep_results_ = pep_results

        return self

    def _add_invalid_pep_reason(
        self,
        pep_results: pd.DataFrame,
        q_peak_samples: pd.DataFrame,
        b_point_samples: pd.DataFrame,
    ) -> pd.DataFrame:

        # extract nan_reason from q_peak_samples and add to pep_results
        pep_results = pep_results.assign(nan_reason=q_peak_samples["nan_reason"])
        # TODO add option to store multiple nan_reasons in one column?
        # extract nan_reason from b_point_samples
        nan_reason_b_point = b_point_samples["nan_reason"].loc[~b_point_samples["nan_reason"].isna()]
        # add nan_reason to pep_results
        if not nan_reason_b_point.empty:
            pep_results.loc[nan_reason_b_point.index, "nan_reason"] = nan_reason_b_point

        neg_pep_idx = pep_results["pep_ms"] < 0
        if self.handle_negative_pep == "zero":
            pep_results.loc[neg_pep_idx, ["pep_sample", "pep_ms"]] = 0
            pep_results.loc[neg_pep_idx, "nan_reason"] = "negative_pep"
        elif self.handle_negative_pep == "nan":
            pep_results.loc[neg_pep_idx, ["pep_sample", "pep_ms"]] = pd.NA
            pep_results.loc[neg_pep_idx, "nan_reason"] = "negative_pep"

        return pep_results



from typing import Union

import pandas as pd
from tpcp import Algorithm

__all__ = ["BaseHeartbeatSegmentation"]


class BaseHeartbeatSegmentation(Algorithm):
    """Base class for all heartbeat segmentation algorithms."""

    _action_methods = "extract"

    # result
    heartbeat_list_: pd.DataFrame

    def extract(
        self,
        *,
        ecg: Union[pd.Series, pd.DataFrame],
        sampling_rate_hz: int,
    ):
        """Segment ECG signal into heartbeats."""
        raise NotImplementedError("Method 'extract' must be implemented in subclass.")



from typing import Optional

import pandas as pd

from biopsykit.signals._base_extraction import BaseExtraction

__all__ = ["BaseCPointExtraction"]


class BaseCPointExtraction(BaseExtraction):
    def extract(self, *, icg: pd.Series, heartbeats: pd.DataFrame, sampling_rate_hz: Optional[float]):
        raise NotImplementedError("This is an abstract method and needs to be implemented in a subclass.")



import pandas as pd

__all__ = ["BaseEcgExtraction"]

from biopsykit.signals._base_extraction import BaseExtraction


class BaseEcgExtraction(BaseExtraction):
    def extract(
        self,
        *,
        ecg: pd.Series,
        heartbeats: pd.DataFrame,
        sampling_rate_hz: float,
    ):
        raise NotImplementedError("This is an abstract method and needs to be implemented in a subclass.")



"""Module for ECG segmentation."""
from biopsykit.signals.ecg.segmentation._base_segmentation import BaseHeartbeatSegmentation
from biopsykit.signals.ecg.segmentation._heartbeat_segmentation import HeartbeatSegmentationNeurokit

__all__ = ["BaseHeartbeatSegmentation", "HeartbeatSegmentationNeurokit"]



import warnings
from typing import Optional, Union

import numpy as np
import pandas as pd
from scipy.signal import argrelmin
from tpcp import Parameter

from biopsykit.signals._base_extraction import HANDLE_MISSING_EVENTS, CanHandleMissingEventsMixin
from biopsykit.signals._dtypes import assert_sample_columns_int
from biopsykit.signals.icg.event_extraction._base_b_point_extraction import BaseBPointExtraction
from biopsykit.utils._datatype_validation_helper import _assert_has_columns, _assert_is_dtype
from biopsykit.utils.array_handling import sanitize_input_dataframe_1d
from biopsykit.utils.exceptions import EventExtractionError

__all__ = ["BPointExtractionForouzanfar2018"]


class BPointExtractionForouzanfar2018(BaseBPointExtraction, CanHandleMissingEventsMixin):
    """algorithm to extract B-point based on [Forouzanfar et al., 2018, Psychophysiology]."""

    # input parameters
    scaling_factor: Parameter[float]
    correct_outliers: Parameter[bool]

    def __init__(
        self,
        scaling_factor: float = 2000,
        correct_outliers: Optional[bool] = False,
        handle_missing_events: HANDLE_MISSING_EVENTS = "warn",
    ):
        """Initialize new BPointExtractionForouzanfar algorithm instance.

        WARNING: In the original paper, the authors report the sampling frequency of the ICG signal as the scaling
        factor. Since this does not make sense, the scaling factor is set to 2000 (corresponding to a sampling rate of
        the original data of 2000 Hz) instead of using the sampling rate of the ICG signal.

        Parameters
        ----------
        scaling_factor : float
            Scaling factor for the B-point extraction algorithm
        correct_outliers : bool
            Indicates whether to perform outlier correction (True) or not (False)
        """
        super().__init__(handle_missing_events=handle_missing_events)
        self.scaling_factor = scaling_factor
        self.correct_outliers = correct_outliers

    # @make_action_safe
    def extract(  # noqa: PLR0915
        self,
        *,
        icg: Union[pd.Series, pd.DataFrame],
        heartbeats: pd.DataFrame,
        c_points: pd.DataFrame,
        sampling_rate_hz: float,  # noqa: ARG002
    ):
        """Extract B-points from given ICG cleaned signal.

        The results are saved in the 'points_' attribute of the class instance.

        Parameters
        ----------
        icg : :class:`~pandas.DataFrame`
            cleaned ICG signal
        heartbeats : :class:`~pandas.DataFrame`
            DataFrame containing one row per segmented heartbeat, each row contains start, end, and R-peak location
            (in samples from beginning of signal) of that heartbeat, index functions as id of heartbeat
        c_points : :class:`~pandas.DataFrame`
            DataFrame containing one row per segmented C-point, each row contains location
            (in samples from beginning of signal) of that C-point or NaN if the location of that C-point
            is not correct.
        sampling_rate_hz : int
            sampling rate of ECG signal in hz

        Returns
        -------
        self

        Raises
        ------
        :exc:`~biopsykit.utils.exceptions.EventExtractionError`
            If the C-Point contains NaN values and handle_missing is set to "raise"

        """
        # sanitize input signal
        self._check_valid_missing_handling()
        icg = sanitize_input_dataframe_1d(icg, column="icg_der")
        icg = icg.squeeze()

        # Create the B-Point/A-Point Dataframes with the index of the heartbeat_list
        b_points = pd.DataFrame(index=heartbeats.index, columns=["b_point_sample", "nan_reason"])

        # check whether the c_points contain NaN
        check_c_points = pd.isna(c_points["c_point_sample"]).to_numpy()

        # Calculate the second- and third-derivative of the ICG-signal
        second_der = np.gradient(icg)
        third_der = np.gradient(second_der)

        for idx, data in heartbeats[1:].iterrows():
            # check if the current or the previous C-Point contain NaN . If this is the case, set the b_point to NaN
            if check_c_points[idx] or check_c_points[idx - 1]:
                b_points["b_point_sample"].iloc[idx] = np.nan
                b_points["nan_reason"].iloc[idx] = "c_point_nan"
                continue

            # Detect the main peak in the dZ/dt signal (C-Point)
            c_point = c_points["c_point_sample"].iloc[idx]

            # Compute the beat to beat interval
            beat_to_beat = c_points["c_point_sample"].iloc[idx] - c_points["c_point_sample"].iloc[idx - 1]

            # Compute the search interval for the A-Point
            search_interval = int(beat_to_beat / 3)

            # Detect the local minimum (A-Point) within one third of the beat to beat interval prior to the C-Point
            a_point = self._get_a_point(icg, search_interval, c_point) + (c_point - search_interval)

            # Select the signal_segment between the A-Point and the C-Point
            signal_clean_segment = icg.iloc[a_point : c_point + 1]

            # Define the C-Point amplitude which is used as a constraint for monotonic segment detection
            c_amplitude = icg.iloc[c_point]

            # Step 4.1: Get the most prominent monotonic increasing segment between the A-Point and the C-Point
            start_sample, end_sample = (
                self._get_most_prominent_monotonic_increasing_segment(signal_clean_segment, c_amplitude) + a_point
            )

            if (start_sample == a_point) & (end_sample == a_point):
                if self.correct_outliers:
                    b_points["b_point_sample"].iloc[idx] = data["r_peak_sample"]
                else:
                    b_points["b_point_sample"].iloc[idx] = np.nan
                    b_points["nan_reason"].iloc[idx] = "no_monotonic_segment"
                continue

            # Get the first third of the monotonic increasing segment
            start = start_sample
            end = end_sample - int((2 / 3) * (end_sample - start_sample))

            # 2nd derivative of the segment
            monotonic_segment_2nd_der = pd.DataFrame(second_der[start:end], columns=["2nd_der"])
            # 3rd derivative of the segment
            monotonic_segment_3rd_der = pd.DataFrame(third_der[start:end], columns=["3rd_der"])

            # Calculate the amplitude difference between the C-Point and the A-Point
            height = icg.iloc[c_point] - icg.iloc[a_point]

            # Compute the significant zero_crossings
            significant_zero_crossings = self._get_zero_crossings_3rd_derivative(
                monotonic_segment_3rd_der, monotonic_segment_2nd_der, height
            )

            # Compute the significant local maxima of the 3rd derivative of the most prominent monotonic segment
            significant_local_maxima = self._get_local_maxima_3rd_derivative(monotonic_segment_3rd_der, height)

            # Label the last zero crossing/ local maximum as the B-Point
            # If there are no zero crossings or local maxima use the first Point of the segment as B-Point
            significant_features = pd.concat([significant_zero_crossings, significant_local_maxima], axis=0) + start
            if len(significant_features) == 0:
                b_point = start
            else:
                b_point = significant_features.iloc[np.argmin(c_point - significant_features)][0]

            b_points["b_point_sample"].iloc[idx] = b_point

        # interpolate the first B-Point with the second B-Point since it is not possible to detect a B-Point
        # for the first heartbeat
        b_points["b_point_sample"].iloc[0] = (
            b_points["b_point_sample"].iloc[1] - b_points["b_point_sample"].diff().iloc[2]
        )

        idx_nan = b_points["b_point_sample"].isna()
        if idx_nan.sum() > 0:
            idx_nan = list(b_points.index[idx_nan])
            missing_str = (
                f"Either 'r_peak' or 'c_point' contains NaN at positions {idx_nan}! The B-points were set to NaN."
            )
            if self.handle_missing_events == "warn":
                warnings.warn(missing_str)
            elif self.handle_missing_events == "raise":
                raise EventExtractionError(missing_str)

        _assert_is_dtype(b_points, pd.DataFrame)
        _assert_has_columns(b_points, [["b_point_sample", "nan_reason"]])
        b_points = b_points.astype({"b_point_sample": "Int64", "nan_reason": "object"})
        assert_sample_columns_int(b_points)

        self.points_ = b_points
        return self

    @staticmethod
    def _get_a_point(icg: pd.Series, search_interval: int, c_point: int):
        signal_interval = icg.iloc[(c_point - search_interval) : c_point]
        signal_minima = argrelmin(signal_interval.values, mode="wrap")

        a_point_candidates = signal_interval.iloc[signal_minima[0]]
        if len(a_point_candidates) == 0:
            # no local minima found => return the argmin of the interval
            return np.argmin(signal_interval)

        a_point_idx = np.argmin(a_point_candidates)
        a_point = signal_minima[0][a_point_idx]
        return a_point

    @staticmethod
    def _get_most_prominent_monotonic_increasing_segment(icg_segment: pd.Series, height: int):
        icg_segment = icg_segment.copy()
        icg_2nd_der_segment = np.gradient(icg_segment)
        icg_segment.index = np.arange(0, len(icg_segment))
        monotony_df = pd.DataFrame(icg_segment.values, columns=["icg"])
        monotony_df = monotony_df.assign(**{"2nd_der": icg_2nd_der_segment, "borders": 0})

        # A-Point is a possible start of the monotonic segment
        monotony_df.loc[monotony_df.index[0], "borders"] = "start_increase"
        # C-Point is a possible end of the monotonic segment
        monotony_df.loc[monotony_df.index[-1], "borders"] = "end_increase"

        neg_pos_change_idx = np.where(np.diff(np.sign(monotony_df["2nd_der"])) > 0)[0]
        monotony_df.loc[monotony_df.index[neg_pos_change_idx], "borders"] = "start_increase"

        pos_neg_change_idx = np.where(np.diff(np.sign(monotony_df["2nd_der"])) < 0)[0]
        monotony_df.loc[monotony_df.index[pos_neg_change_idx], "borders"] = "end_increase"

        # drop all samples that are no possible start-/ end-points
        monotony_df = monotony_df.drop(monotony_df[monotony_df["borders"] == 0].index)
        monotony_df = monotony_df.reset_index()

        # Drop start- and corresponding end-point if their start value is higher than 1/2 of H
        start_index_drop_rule_a = monotony_df[
            (monotony_df["borders"] == "start_increase") & (monotony_df["icg"] > height / 2)
        ].index
        start_index_drop_rule_a = start_index_drop_rule_a.union(start_index_drop_rule_a + 1)
        monotony_df = monotony_df.drop(index=start_index_drop_rule_a)

        # Drop start- and corresponding end-point if their end values does not reach at least 2/3 of H
        end_index_drop_rule_b = monotony_df[
            (monotony_df["borders"] == "end_increase") & (monotony_df["icg"] < 2 * height / 3)
        ].index

        end_index_drop_rule_b = end_index_drop_rule_b.union(end_index_drop_rule_b - 1)
        monotony_df = monotony_df.drop(index=end_index_drop_rule_b)

        # Select the monotonic segment with the highest amplitude difference
        start_sample = 0
        end_sample = 0
        if len(monotony_df) > 2:
            idx = np.argmax(monotony_df["icg"].diff())
            start_sample = monotony_df["index"].iloc[idx - 1]
            end_sample = monotony_df["index"].iloc[idx]
        elif len(monotony_df) != 0:
            start_sample = monotony_df["index"].iloc[0]
            end_sample = monotony_df["index"].iloc[-1]

        return start_sample, end_sample  # That are not absolute positions yet

    def _get_zero_crossings_3rd_derivative(
        self, monotonic_segment_3rd_der: pd.DataFrame, monotonic_segment_2nd_der: pd.DataFrame, height: int
    ):
        constraint = float(10 * height / self.scaling_factor)

        zero_crossings = np.where(np.diff(np.signbit(monotonic_segment_3rd_der["3rd_der"])))[0]
        zero_crossings = pd.DataFrame(zero_crossings, columns=["sample_position"])

        # Discard zero_crossings with negative to positive sign change
        significant_crossings = zero_crossings.drop(
            zero_crossings[monotonic_segment_2nd_der.iloc[zero_crossings["sample_position"]].to_numpy() < 0].index,
            axis=0,
        )

        # Discard zero crossings with slope higher than 10*H/scaling_factor
        significant_crossings = significant_crossings.drop(
            significant_crossings[
                monotonic_segment_2nd_der.iloc[significant_crossings["sample_position"]].to_numpy() >= constraint
            ].index,
            axis=0,
        )

        if isinstance(zero_crossings, type(None)) or len(zero_crossings) == 0:
            return pd.DataFrame([0], columns=["sample_position"])
        return significant_crossings

    def _get_local_maxima_3rd_derivative(self, monotonic_segment_3rd_der: pd.DataFrame, height: int):
        constraint = float(4 * height / self.scaling_factor)
        # compute gradient
        monotonic_segment_3rd_der_gradient = np.gradient(monotonic_segment_3rd_der.squeeze())

        # find zero-crossings of the gradient to determine local maxima
        local_maxima = np.where(np.diff(np.sign(monotonic_segment_3rd_der_gradient)) < 0)[0]
        # add 1 to the index to get the correct position
        local_maxima += 1
        local_maxima = pd.DataFrame(local_maxima, columns=["sample_position"])

        significant_maxima = local_maxima.drop(
            local_maxima[
                monotonic_segment_3rd_der["3rd_der"].iloc[local_maxima["sample_position"]].to_numpy() < constraint
            ].index,
            axis=0,
        )

        if isinstance(significant_maxima, type(None)):
            return pd.DataFrame([0], columns=["sample_position"])
        if len(significant_maxima) == 0:
            return pd.DataFrame([0], columns=["sample_position"])
        return significant_maxima



from typing import Optional

import pandas as pd

from biopsykit.signals._dtypes import assert_sample_columns_int
from biopsykit.signals.icg.outlier_correction import BaseOutlierCorrection

__all__ = ["OutlierCorrectionDummy"]

from biopsykit.utils._datatype_validation_helper import _assert_has_columns, _assert_is_dtype


class OutlierCorrectionDummy(BaseOutlierCorrection):
    """Dummy class for outlier correction. Does nothing and passes through the input data unchanged."""

    def correct_outlier(
        self,
        *,
        b_points: pd.DataFrame,
        c_points: Optional[pd.DataFrame],  # noqa: ARG002
        sampling_rate_hz: Optional[float],  # noqa: ARG002
        **kwargs,  # noqa: ARG002
    ):
        """Correct no outliers, just pass through the input data unchanged.

        The results of the outlier correction are saved in the `points_` attribute of the class instance.

        Parameters
        ----------
        b_points : :class:`~pandas.DataFrame`
            Dataframe containing the extracted B-Points per heartbeat, index functions as id of heartbeat
        c_points: :class:`~pandas.DataFrame`
            Dataframe containing the extracted C-Points per heartbeat, index functions as id of heartbeat. Not used.
        sampling_rate_hz : int
            Sampling rate of ICG signal in hz. Not used.
        kwargs: dict
            Additional keyword arguments. Not used.

        Returns
        -------
        self

        """
        _assert_is_dtype(b_points, pd.DataFrame)
        _assert_has_columns(b_points, [["b_point_sample", "nan_reason"]])
        b_points = b_points.astype({"b_point_sample": "Int64", "nan_reason": "object"})
        assert_sample_columns_int(b_points)

        self.points_ = b_points
        return self



from typing import Optional

import pandas as pd
from scipy import signal


def clean_icg_deriv(
    raw_signal: pd.Series, sampling_rate_hz: int, filter_type: Optional[str] = "butterworth"
) -> pd.Series:
    """Clean ICG dZ/dt signal using a band-pass filter.

    This function filters the raw dZ/dt ICG signal using a band-pass filter. The following filters are available:
        * Butterworth band-pass filter: 4th order, low cutoff 0.5 Hz, high cutoff 25 Hz, see Forouzanfar 2019
        * Elliptic band-pass filter (Cauer filter): 2nd order, low cutoff 0.75 Hz, high cutoff 40 Hz, pass-band ripple
            1 dB, stop-band ripple 80 dB, see Nabian 2017 & Ostadabbas 2023
        * Savitzky-Golay filter: low-pass, 3rd order polynomial, see Salah 2020 & Salah 2017 (only useful for signal
            with high frequency noise?)

    Parameters
    ----------
    raw_signal: pd.Series
        pd.Series containing the raw dZ/dt ICG signal
    sampling_rate_hz: int
        sampling rate of ICG dZ/dt signal in hz
    filter_type: str, optional
        type of filter. Must be one of "butterworth", "elliptic", or "savgol". Default: "butterworth"

    Returns
    -------
    pd.Series
        cleaned ICG dZ/dt signal

    """
    if filter_type not in ["butterworth", "elliptic", "savgol"]:
        raise ValueError("Filter type can only be 'butterworth', 'elliptic', or 'savgol'")

    if filter_type == "butterworth":
        sos = signal.butter(N=4, Wn=[0.5, 25], btype="bandpass", output="sos", fs=sampling_rate_hz)
        clean_signal = signal.sosfiltfilt(sos, raw_signal)

    elif filter_type == "elliptic":
        rp = 1.0
        rs = 80.0
        sos = signal.ellip(
            N=2, rp=rp, rs=rs, Wn=[0.75, 40], btype="bandpass", output="sos", fs=sampling_rate_hz, analog=False
        )
        clean_signal = signal.sosfiltfilt(sos, raw_signal)

    elif filter_type == "savgol":  # Savitzky-Golay filter (for high frequency noise?!)
        clean_signal = signal.savgol_filter(raw_signal, window_length=61, polyorder=3)

    else:
        raise ValueError("Filter type can only be 'butterworth', 'elliptic', or 'savgol'")

    clean_signal = pd.Series(clean_signal, index=raw_signal.index, name="icg_der")
    return clean_signal



from typing import Optional

import neurokit2 as nk
import pandas as pd


def clean_ecg(raw_signal: pd.Series, sampling_rate_hz: int, method: Optional[str] = "biosppy") -> pd.Series:
    """Clean ECG signals using :func:`~neurokit2.ecg_clean`.

    Args:
        raw_signal: pd.Series containing the raw ECG signal
        sampling_rate_hz: sampling rate of the ECG signal in hz
        method: cleaning method (default is "biosppy"), can be any either "neurokit" or "biosppy"

    Returns
    -------
        clean_signal: pd.DataFrame containing filtered signal while keeping index of input signal
    """
    if method not in ["neurokit", "biosppy"]:
        raise ValueError("Not implemented yet!")

    clean_signal = nk.ecg_clean(raw_signal, sampling_rate=sampling_rate_hz, method=method)
    clean_signal = pd.Series(clean_signal, index=raw_signal.index, name="ecg")

    return clean_signal



import warnings
from typing import Optional

import numpy as np
import pandas as pd

from biopsykit.signals._base_extraction import HANDLE_MISSING_EVENTS, CanHandleMissingEventsMixin
from biopsykit.signals._dtypes import assert_sample_columns_int
from biopsykit.signals.icg.event_extraction._base_b_point_extraction import BaseBPointExtraction
from biopsykit.utils._datatype_validation_helper import _assert_has_columns, _assert_is_dtype
from biopsykit.utils.array_handling import sanitize_input_dataframe_1d
from biopsykit.utils.exceptions import EventExtractionError


class BPointExtractionSherwood1990(BaseBPointExtraction, CanHandleMissingEventsMixin):
    def __init__(self, handle_missing_events: HANDLE_MISSING_EVENTS = "warn"):
        """Initialize new BPointExtractionSherwood1990 algorithm instance."""
        super().__init__(handle_missing_events=handle_missing_events)

    def extract(
        self,
        *,
        icg: pd.Series,
        heartbeats: pd.DataFrame,
        c_points: pd.DataFrame,
        sampling_rate_hz: Optional[float],  # noqa: ARG002
    ):
        self._check_valid_missing_handling()
        # sanitize input
        icg = sanitize_input_dataframe_1d(icg, column="icg_der")
        icg = icg.squeeze()

        # Create the b_point Dataframe. Use the heartbeats id as index
        b_points = pd.DataFrame(index=heartbeats.index, columns=["b_point_sample", "nan_reason"])

        # get the c_point locations from the c_points dataframe and search for entries containing NaN
        c_points = c_points["c_point_sample"]
        check_c_points = pd.isna(c_points)

        # get zero crossings of icg
        zero_crossings = np.where(np.diff(np.signbit(icg)))[0]

        # go through each R-C interval independently and search for the local minima
        for idx, data in heartbeats.iterrows():
            # check if r_peaks/c_points contain NaN. If this is the case, set the b_point to NaN and continue
            # with the next iteration
            if check_c_points[idx]:
                b_points["b_point_sample"].iloc[idx] = np.NaN
                b_points["nan_reason"].iloc[idx] = "c_point_nan"
                missing_str = f"The c_point contains NaN at position {idx}! B-Point was set to NaN."
                if self.handle_missing_events == "warn":
                    warnings.warn(missing_str)
                elif self.handle_missing_events == "raise":
                    raise EventExtractionError(missing_str)
                continue

            # get the closest zero crossing *before* the C-point
            c_point = c_points[idx]
            zero_crossings_diff = zero_crossings - c_point
            zero_crossings_diff = zero_crossings_diff[zero_crossings_diff < 0]
            zero_crossing_idx = np.argmax(zero_crossings_diff)

            b_point = zero_crossings[zero_crossing_idx]
            # assert that b_point is within the R-C interval
            if not (data["r_peak_sample"] < b_point < c_point):
                b_point = np.NaN
                b_points["nan_reason"].iloc[idx] = "no_zero_crossing"

            # Add the detected B-point to the b_points Dataframe
            b_points["b_point_sample"].iloc[idx] = b_point

        _assert_is_dtype(b_points, pd.DataFrame)
        _assert_has_columns(b_points, [["b_point_sample", "nan_reason"]])
        b_points = b_points.astype({"b_point_sample": "Int64", "nan_reason": "object"})
        assert_sample_columns_int(b_points)

        self.points_ = b_points

        return self



import pandas as pd

from biopsykit.utils.exceptions import ValidationError

__all__ = ["assert_sample_columns_int"]


def assert_sample_columns_int(data: pd.DataFrame) -> None:
    """Assert that the columns of a DataFrame that have "_sample" in their name are of type int."""
    if not any(data.columns.str.contains("_sample")):
        raise ValidationError("DataFrame does not contain any columns with '_sample' in their name!")
    for col in data.columns:
        if "_sample" in col and not pd.api.types.is_integer_dtype(data[col]):
            raise ValidationError(f"Column '{col}' is not of type 'int'!")
